/*
 * getcdlabel (early Alpha version)
 *
 * Prints the label of a data cdrom.
 *
 * Most code was ripped and modified from Wine (cdrom.c) and
 * Unicode support from Linux kernel 2.4.
 *
 * Copyright © 2001 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>

#include "diags.h"
#include "nls.h"


#define VOLDESC_SET_TERMINATOR	0xff

static int
cdrom_find_best_voldesc(int fd)
{
	unsigned char cur_vd_type, max_vd_type = 0;
	unsigned int offs, best_offs = 0;

	for (offs = 0x8000; offs <= 0x9800; offs += 0x800) {
		lseek(fd, offs, SEEK_SET);
		read(fd, &cur_vd_type, 1);

		if (cur_vd_type == VOLDESC_SET_TERMINATOR)
			break;
		if (cur_vd_type > max_vd_type) {
			max_vd_type = cur_vd_type;
			best_offs = offs;
		}
	}

	return best_offs;
}

#define LABEL_LEN	32 + 1

static int
cdrom_getlabel(int dev, char *label)
{
	int offs = cdrom_find_best_voldesc(dev);
	wchar_t label_read[LABEL_LEN]; /* Unicode possible, too */
	int unicode_id = 0;

	if (offs) {
		if ((lseek(dev, offs + 0x58, SEEK_SET) == offs + 0x58) &&
		    (read(dev, &unicode_id, 3) == 3)) {
			int ver = (unicode_id & 0xff0000) >> 16;

			if ((lseek(dev, offs + 0x28, SEEK_SET) != offs + 0x28) ||
			    (read(dev, &label_read, LABEL_LEN) != LABEL_LEN))
				die(10, "reading label !");

			if (((unicode_id & 0xffff) == 0x2f25) &&
			    ((ver == 0x40) || (ver == 0x43) || (ver == 0x45))) {
				int i;

				for (i = 0; i < LABEL_LEN; i++)
					uni2char(label_read[i], &label[i]);

				label[LABEL_LEN] = '\0';
			} else {
				strncpy(label, (char *)label_read, 11);
				label[11] = '\0';
			}

			return 1;
		}
	}

	return 0;
}

int
main(int argc, char **argv)
{
	int cdrom, type;
	char label[LABEL_LEN], *name = NULL;
	char *path = *(argv + 1);

	if (argc < 2)
		die(1, "missing path argument");

	cdrom = open(path, O_RDONLY | O_NONBLOCK);
	if (cdrom == -1)
		die(10, "can't open cdrom device (%s)", path);

	type = ioctl(cdrom, CDROM_DISC_STATUS);

	switch (type) {
	case CDS_AUDIO:
		name = "Audio";
		strcpy(label, "Audio CD   ");
		break;

	/* Fall through for all data CD types. */
	case CDS_DATA_1:
		if (!name)
			name = "Data_1";
	case CDS_DATA_2:
		if (!name)
			name = "Data_2";
	case CDS_XA_2_1:
		if (!name)
			name = "XA 2.1";
	case CDS_XA_2_2:
		if (!name)
			name = "XA 2.2";
	case -1:
		if (!name)
			name = "Unknown/ISO file";

		/* Common code here. Hopefully a data CD. */
		cdrom_getlabel(cdrom, label);
		break;

	case CDS_MIXED:
		name = "Mixed mode";
		warning("can't retrieve the label of a mixed mode CD");
		/* Fall through. */

	case CDS_NO_INFO:
		if (!name)
			name = "No_info";
		break;

	default:
		warning("Strange CD type (%d) or empty ?\n", type);
		name = "Strange/empty";
		break;
	}

	puts(label);
	fprintf(stderr, "Info: cdtype = %s\n", name);

	close(cdrom);

	return 0;
}
